function [predsAfterStage1,predsAfterStage2,modelsAfterStage2] =...
    SelfSupAM_crossValidation( varargin )
%   SelfSupAM_crossValidation Evaluates a Self-Supervised ARTMAP voting system 
%	via cross-validation on a benchmark dataset using the Java class
%   edu.bu.cns.classifiers.SelfSupCrossValidator
%
%   [predsAfterStage1,predsAfterStage2,modelsAfterStage2]
%   = SelfSupAM_crossValidation( strips, stage1FeatureIndexes ) 
%     Trains and tests a VoterEnsemble of SelfSupervisedARTMAPs 
%       with default parameter values, using CV on the passed strips
%       strips ~ 1xF struct of benchmark strips with fields "inputs" and "labels"
%       stage1FeatureIndexes ~ 1xMbar list of feature indexes to be
%           included in Stage 1 training.
%       predsAfterStage1,predsAfterStage2 ~ 1xF cell arrays of label predictions 
%           after Stage 1 and after Stage 2 training
%       modelsAfterStage2 ~ 1xF cell of fully trained models
%
%   [...] = SelfSupAM_crossValidation( strips, stage1FeatureIndexes, "PropertyName1",propValueel,... ) 
%     Same as above, but using the passed model parameters.
%
%   [...] = SelfSupAM_crossValidation( strips, stage1FeatureIndexes, model ) 
%     Same as above, but using the passed model
%       model ~ Java model to clone and use for CV
%
%           
%   Based on:   Amis, G., & Carpenter, G.A. (2009). 
%                   Self-Supervised ARTMAP. (in preparation)
%   
%
%   Greg Amis
%   Technology Lab
%   Dept. of Cognitive & Neural Systems
%   Boston University
%   29 March 2009
%
%   Licensing Policy
%   This implementation is made available to the general public under the terms 
%   of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
%   (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
%   As such, it is available free of charge, and may be copied, redistributed, or modified, 
%   as long as any resulting work is distributed under the same terms of Copyleft.

if nargin<2, error('At least the strips and the stage1FeatureIndexes are required.'); end;

SelfSupAM_loadJava();

strips = varargin{1};
stage1FeatureIndexes = varargin{2};

if nargin==3
    model = varargin{3};
else
    %Create model with default parameters
    params = {'V',5};
    if nargin>3
        params = [params,varargin(3:end)];
    end;
    model = SelfSupAM_createInstance(params{:});
end;
    
M = model.getNumberOfFeatures();
if M==-1
    M = size(strips(1).inputs,1);
    model.setNumberOfFeatures(M);
end;

L = model.getNumberOfClasses();
if L==-1
    L = max(strips(1).labels); 
    model.setNumberOfClasses(L);
end;

F = length(strips);
jstrips = javaArray('edu.bu.cns.data.LabeledInputs',F);
for f = 1:F
    labels = int32(strips(f).labels);
    isLabeled = labels~=-1;
    labels(isLabeled) = labels(isLabeled)-1; 
    jstrips(f) = edu.bu.cns.data.LabeledInputsImpl(...
        strips(f).inputs', int32(labels), L);
end;

cv = edu.bu.cns.classifiers.SelfSupCrossValidator( model, jstrips, int32(stage1FeatureIndexes-1) );

results = cv.call();

predsAfterStage1 = cell(1,F);
predsAfterStage2 = cell(1,F);
modelsAfterStage2 = cell(1,F);

for f = 1:F
    predsAfterStage1{f} = results.predsAfterStage1.get(f-1)' + 1;
    predsAfterStage2{f} = results.predsAfterStage2.get(f-1)' + 1;
    modelsAfterStage2{f} = results.modelsAfterStage2.get(f-1);
end;

