function [inputs, labels] = create_diagnosis_data(N,overlap)
% create_diagnosis_data  Creates training data for 2D medical diagnosis
%   illustration, as described in:
%       
%       Amis, G., & Carpenter, G.A. (2009). Self-Supervised ARTMAP. (in preparation)
%
%   [inputs, labels] = create_diagnosis_data(N,overlap)
%       Returns 2xN inputs and their N associated class labels, 
%           overlap ~ scalar between 0 and 1
%               the fraction by which the hypothermic and septic clusters
%               overlap the normal cluster in dimension 1 (temperature)
%
%   [inputs, labels] = create_diagnosis_data(N)
%       Uses the default overlap=0.7.
%
%   Greg Amis
%   Technology Lab
%   Dept. of Cognitive & Neural Systems
%   Boston University
%   27 March 2009
%
%   Licensing Policy
%   This implementation is made available to the general public under the terms 
%   of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
%   (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
%   As such, it is available free of charge, and may be copied, redistributed, or modified, 
%   as long as any resulting work is distributed under the same terms of
%   Copyleft.

if nargin==1, overlap = 0.7; end;

padding = 0.2; %minimum relative distance between clusters and feature space edge
scaling = [3;2.6]; %to scale inputs to unit square.

%Lower-left point of each cluster
septicStart = [ 1+padding+(1-overlap)*2; 1+padding*2 ]; 
hypoStart = [ padding; 1+padding*2 ];
normalStart = [ padding+1-overlap; padding ];
    
Nsick = round((N/4)); %Number of points in each sick cluster
Nnormal = N - 2*Nsick; %Twice as many normal points as sick
    %This creates the same point density in each cluster,
    %  since normal cluster is twice as wide as sick clusters.

onesSick = ones(1,Nsick);
onesNormal = ones(1,Nnormal);

%Inputs for each class
aSeptic = septicStart(:,onesSick) + rand(Nsick,2)';
aHypo = hypoStart(:,onesSick) + rand(Nsick,2)';
aNormal = normalStart(:,onesNormal) + rand(Nnormal,2)'.*([2;1]*onesNormal);

inputs = [ aSeptic, aHypo, aNormal ] ./ scaling(:,ones(1,N));
Hypo = 2; Normal = 3; Septic = 1;
labels = [ Septic(:,onesSick), Hypo(:,onesSick), Normal(:,onesNormal) ];

%scramble
randOrder = randperm(N);
inputs = inputs(:,randOrder);
labels = labels(randOrder);

return;
    
%% Test it out
[inputs,labels] = create_diagnosis_data(1000); %#ok<UNRCH>
colors = 'rbg';
figure; hold on;
for k = 1:3
    isk = labels==k;
    scatter(inputs(1,isk),inputs(2,isk),[colors(k),'.']);
end;
axis square; axis([ 0 1 0 1]);
    