/**
 *	VoterEnsemble.java
 *
 *  Greg Amis
 *  Technology Lab
 *  Dept. of Cognitive & Neural Systems
 *  Boston University
 *  Apr 11, 2008
 *
 *  Licensing Policy
 *  This implementation is made available to the general public under the terms 
 *  of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
 *  (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
 *  As such, it is available free of charge, and may be copied, redistributed, or modified, 
 *  as long as any resulting work is distributed under the same terms of Copyleft.
 */
package edu.bu.cns.classifiers;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.concurrent.Callable;

import edu.bu.cns.data.Input;
import edu.bu.cns.data.Inputs;
import edu.bu.cns.data.LabeledInputs;
import edu.bu.cns.util.DeepCopier;
import edu.bu.cns.util.ForkProcessor;
import edu.bu.cns.util.IntArrayList;

/**
 *	Multithreaded implementation. Voters train and test in parallel,
 *  to the degree supported by the hardware.
 *  class prediction K' = class with the most votes.
 */
public class VoterEnsemble implements SelfSupervisedClassifier, Cloneable, Serializable {
	
	private static final long serialVersionUID = 3794439541188773721L;
	
	/** Number of voters */
	protected int V = 5;
	
	/** Max number of samples per class */
	protected int P = Integer.MAX_VALUE;
	
	/** cloned to produce voters */
	private SelfSupervisedClassifier template;
	
	private List<SelfSupervisedClassifier> voters;
	
	private long tieBreakerSeed = 1l, samplerSeed = 1l;
	
	private transient IntArrayList ties;
	private transient Random tieBreaker, sampler;
	private transient ForkProcessor<int[]> forkProcessor;
	
	public void setUp() {
		if (voters==null) {
			if (template==null)
				throw new IllegalParameterException("Voter template not set");
			voters = new ArrayList<SelfSupervisedClassifier>(V);
			for (int v=0; v<V; v++) {
				SelfSupervisedClassifier voter = template.clone();
				voters.add(voter);
			}
		}
		for (final SelfSupervisedClassifier v : voters)
			v.setUp();

		tieBreaker = new Random(tieBreakerSeed);
		ties = new IntArrayList(getNumberOfClasses());
		sampler = new Random(samplerSeed);
		forkProcessor = new ForkProcessor<int[]>(V);
	}
	
	/**
	 * By calling trainStage1, trainStage2, and test at the same time, 
	 *   exercise makes better use of multithreading, 
	 *   eliminating unnecessary thread joins.
	 * @param stage1Training	Labeled training data
	 * @param stage2Training	Unlabeled training data 
	 * @param testing	Unlabeled data for testing
	 * @return class label predictions
	 */
	public int[] exercise( 
			final LabeledInputs stage1Training, 
			final Inputs<?> stage2Training, 
			final Inputs<?> testing )
	{
		for (final SelfSupervisedClassifier v : voters) {
			forkProcessor.submit(new Callable<int[]>() { public int[] call() {
					trainStage1(v, stage1Training);
					trainStage2(v, stage2Training);
					return v.test(testing);
				}
			});
		}
		final List<int[]> results = forkProcessor.finish();

		if (forkProcessor.isCanceled()) return null;
		
		return makePredictions(results);
	}
	
	/**
	 * By calling trainStage1, trainStage2, and test at the same time, 
	 *   exerciseDistribution makes better use of multithreading, 
	 *   eliminating unnecessary thread joins.
	 * @param stage1Training	Labeled training data
	 * @param stage2Training	Unlabeled training data 
	 * @param testing	Unlabeled data for testing
	 * @return class label probabilities
	 */
	public float[][] exerciseDistribution( 
			final LabeledInputs stage1Training, 
			final Inputs<?> stage2Training, 
			final Inputs<?> testing )
	{
		for (final SelfSupervisedClassifier v : voters) {
			forkProcessor.submit(new Callable<int[]>() { public int[] call() {
					trainStage1(v, stage1Training);
					trainStage2(v, stage2Training);
					return v.test(testing);
				}
			});
		}
		final List<int[]> results = forkProcessor.finish();

		if (forkProcessor.isCanceled()) return null;
		
		return makeDistributions(results);
	}
	
	/**
	 * By calling train and test at the same time, 
	 *   exercise makes better use of multithreading, 
	 *   eliminating unnecessary thread joins.
	 * @param training	Labeled training data
	 * @param testing	Unlabeled data for testing
	 * @return class label predictions
	 */
	public int[] exercise( 
			final LabeledInputs training, 
			final Inputs<?> testing )
	{
		for (final SupervisedClassifier v : voters) {
			forkProcessor.submit(new Callable<int[]>() { public int[] call() {
					train(v, training);
					return v.test(testing);
				}
			});
		}
		final List<int[]> results = forkProcessor.finish();

		if (forkProcessor.isCanceled()) return null;
		
		return makePredictions(results);
	}
	
	/**
	 * By calling train and test at the same time, 
	 *   exerciseDistribution makes better use of multithreading, 
	 *   eliminating unnecessary thread joins.
	 * @param training	Labeled training data
	 * @param testing	Unlabeled data for testing
	 * @return class label probabilities
	 */
	public float[][] exerciseDistribution( 
			final LabeledInputs training, 
			final Inputs<?> testing )
	{
		for (final SupervisedClassifier v : voters) {
			forkProcessor.submit(new Callable<int[]>() { public int[] call() {
					train(v,training);
					return v.test(testing);
				}
			});
		}
		final List<int[]> results = forkProcessor.finish();

		if (forkProcessor.isCanceled()) return null;
		
		return makeDistributions(results);
	}
	
	/**
	 * Supervised training for a single voter 
	 * 	 on a random subsample of the passed dataset,
	 *   downsampled to class cap.
	 * @param voter
	 * @param src
	 */
	protected void train(final SupervisedClassifier voter, final LabeledInputs src) {
		final LabeledInputs training = src.downsampleToLabelCap(sampler, P);
		voter.train(training);
	}
	
	
	/**
	 * Stage 1 training for a single voter 
	 *   on a random subsample of the passed dataset,
	 *   downsampled to class cap.
	 * @param voter
	 * @param src
	 */
	protected void trainStage1(final SelfSupervisedClassifier voter, final LabeledInputs src) {
		final LabeledInputs stage1Training = src.downsampleToLabelCap(sampler, P);
		voter.trainStage1(stage1Training);
	}
	
	/**
	 * Stage 2 training for a single voter on a shuffled copy of the passed dataset
	 * @param voter
	 * @param src
	 */
	protected void trainStage2(final SelfSupervisedClassifier voter, final Inputs<?> src) {
		final Inputs<?> stage2Training = src.createShuffledCopy(sampler);
		voter.trainStage2(stage2Training);
	}
	
	/**
	 * Predicts the class with the most votes, breaking ties randomly.
	 * @param votes
	 */
	protected int makePrediction(final int[] votes) {
		
		ties.clear();
		
		int winner = -1;
		final int L = getNumberOfClasses();
		
		for (int k = 0; k<L; k++) {
			if (winner==-1 || votes[k]>votes[winner]) {
				winner = k;
				ties.clear();
			}
			else if ( votes[k]==votes[winner] ) {
				if (ties.isEmpty()) 
					ties.add(winner);
				ties.add(k);
			}
		}
		
		if (votes[winner]==0)
			winner = -1;
		else if (ties.size()>1)
			winner = ties.get(tieBreaker.nextInt(ties.size()));
		
		return winner;
	}
	
	protected int[] makePredictions(final List<int[]> results) {

		final int N = results.get(0).length;
		final int L = getNumberOfClasses();
		final int[][] votes = new int[N][L];
		for (final int[] vote : results) {
			for (int n = 0; n<N; n++)
				votes[n][vote[n]]++;
		}
		
		final int[] preds = new int[N];
		for (int n = 0; n<N; n++) {
			final int Kp = makePrediction(votes[n]);			
			preds[n] = Kp;
		}
		
		return preds;
	}
	
	public void train( final LabeledInputs inputs ) {
		trainStage1(inputs);
	}
	
	public void trainStage1( final LabeledInputs inputs ) {
		
		for (final SelfSupervisedClassifier v : voters) {
			forkProcessor.submit(new Callable<int[]>() { public int[] call() {
					trainStage1(v,inputs);
					return null;
				}
			});
		}
		forkProcessor.finish();
	}
	
	public void trainStage2(final Inputs<?> inputs) {
		
		for (final SelfSupervisedClassifier v : voters) {
			forkProcessor.submit(new Callable<int[]>() { public int[] call() {
					trainStage2(v,inputs);
					return null;
				}
			});
		}
		forkProcessor.finish();
	}
	
	public int[] test( final Inputs<?> inputs ) {
		
		
		for (final SelfSupervisedClassifier v : voters) {
			forkProcessor.submit(new Callable<int[]>() { public int[] call() {
					return v.test(inputs);
				}
			});
		}
		
		final List<int[]> results = forkProcessor.finish();

		if (forkProcessor.isCanceled()) return null;
		
		return makePredictions(results);
	}

	public int test(final Input input) {
		final int[] votes = new int[getNumberOfClasses()];
		for (final SupervisedClassifier v : voters) {
			final int Kp = v.test(input);
			votes[Kp]++;
		}
		return makePrediction(votes);
	}

	public float[][] testDistribution( final Inputs<?> inputs ) {
		for (final SupervisedClassifier v : voters) {
			forkProcessor.submit(new Callable<int[]>() { public int[] call() {
					return v.test(inputs);
				}
			});
		}
		
		final List<int[]> results = forkProcessor.finish();

		if (forkProcessor.isCanceled()) return null;

		return makeDistributions(results);
	}

	public float[] testDistribution(final Input input) {
		final float vote = 1f/V;
		final float[] p = new float[getNumberOfClasses()];
		for (final SupervisedClassifier v : voters) {
			final int Kp = v.test(input);
			p[Kp] += vote;
		}
		return p;
	}
	
	protected float[][] makeDistributions(final List<int[]> results)
	{
		final int N = results.get(0).length;
		final int L = getNumberOfClasses();
		final int[][] votes = new int[N][L];
		for (final int[] vote : results) {
			for (int n = 0; n<N; n++)
				votes[n][vote[n]]++;
		}
		
		final float[][] pKs = new float[N][L];
		final float Vf = V;
		for (int n = 0; n<N; n++) {
			final float[] p = pKs[n];
			final int[] vote = votes[n];
			for (int k=0; k<L; k++)
				p[k] = vote[k]/Vf;
		}
		
		return pKs;		
	}
	
	public SelfSupervisedClassifier clone() {
		return DeepCopier.createDeepCopy(this);
	}

	public int getV() {
		return V;
	}

	public void setV(int v) {
		V = v;
	}

	public SelfSupervisedClassifier getTemplate() {
		return template;
	}

	public void setTemplate(SelfSupervisedClassifier template) {
		this.template = template;
	}

	public long getTieBreakerSeed() {
		return tieBreakerSeed;
	}

	public void setTieBreakerSeed(long tieBreakerSeed) {
		this.tieBreakerSeed = tieBreakerSeed;
	}

	public long getSamplerSeed() {
		return samplerSeed;
	}

	public void setSamplerSeed(long samplerSeed) {
		this.samplerSeed = samplerSeed;
	}
	
	public int getClassCap() {
		return P;
	}
	public void setClassCap(int classCap) {
		if (classCap<1) 
			throw new IllegalParameterException("Class cap (p)",classCap,1,Integer.MAX_VALUE);
		this.P = classCap;
	}
	/**
	 * Returns true if a class-cap exists, i.e., P<Integer.MAX_VALUE
	 * @return
	 */
	public boolean hasClassCap() {
		return P<Integer.MAX_VALUE;
	}
	
	public int getNumberOfClasses() { return template.getNumberOfClasses(); }
	public int getNumberOfFeatures() { return template.getNumberOfFeatures(); }
	public void setNumberOfClasses(int L ) { template.setNumberOfClasses(L); }	
	public void setNumberOfFeatures(int M) { template.setNumberOfFeatures(M); }
	
	public List<SelfSupervisedClassifier> voters() {
		return DeepCopier.createDeepCopy(voters);
	}

	@Override
	public String toString() {
		return String.format( "VoterEnsemble(V=%d,P=%d) [ %s ]",
				V, P, template.toString() );
	}


}
