/**
 *	SelfSupervisedARTMAPTest.java
 *
 *  Greg Amis
 *  Technology Lab
 *  Dept. of Cognitive & Neural Systems
 *  Boston University
 *  Mar 24, 2008
 *
 *  Licensing Policy
 *  This implementation is made available to the general public under the terms 
 *  of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
 *  (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
 *  As such, it is available free of charge, and may be copied, redistributed, or modified, 
 *  as long as any resulting work is distributed under the same terms of Copyleft.
 */
package edu.bu.cns.classifiers.artmap;

import static java.lang.Float.NaN;

import java.util.Random;

import junit.framework.AssertionFailedError;
import junit.framework.TestCase;
import edu.bu.cns.classifiers.ClassLabelOutOfBoundsException;
import edu.bu.cns.data.InputFeaturesLengthMismatchException;
import edu.bu.cns.data.LabeledInput;
import edu.bu.cns.data.LabeledInputsImpl;
import edu.bu.cns.util.UnitTestFailureContext;

/**
 *	Unit tests for SelfSupervisedARTMAP, ranging from the most basic ARTMAP functionality
 *	to the illustrations covered in the publication.
 */
public class SelfSupervisedARTMAPTest extends TestCase {
	
	protected SelfSupervisedARTMAP am;
	
	/** 
	 * Most basic, test that an empty artmap creates point box at a. 
	 */
	public void testPointBox() {
		
		am = createInstance(1,2);
		am.setUp();
		
		final float a = 0.25f;
		final int K = 1;
		
		trainStage1( a, K ); //Create point box.
		trainStage1( a, K ); //Just to make sure it doesn't create two boxes.
		
		assertEquals( K, test(a) );
		assertEquals( 1, am.C() );
		assertEquals( a, am.w()[0][0], 1e-6f );
		assertEquals( 1f, am.T()[0], 1e-6f );
	}
	
	
	/**
	 * Exclusive-OR problem.
	 */
	public void testXOR() {
		
		am = createInstance(2,2);
		am.setUp();
		
		final LabeledInputsImpl training = new LabeledInputsImpl(4,2,2);
		training.add( 0,1, 0 );
		training.add( 0,0, 1 );
		training.add( 1,0, 0 );
		training.add( 1,1, 1 );
		
		am.trainStage1( training ); 		//Should create four point boxes.
		am.trainStage1( training.get(0) ); //Shouldn't create any extra node
		
		assertEquals( 4, am.C() );
		
		final int[] preds = am.test(training);
		assertEquals( training.labels(), preds);
		
	}
	
	/**
	 * Makes sure that the system would rather recruit a new node
	 * rather than learn on a node win Tj<Tu.
	 */
	public void testPicksNewOverWorst() {
		
		am = createInstance(2,2);
		am.setUp();
		
		trainStage1(1,1, 0);
		assertEquals( 1, am.C() );
		assertEquals( 1f, am.w()[0][0], 1e-6f );
		assertEquals( 1f, am.w()[0][1], 1e-6f );
		
		test(0,0);
		assertEquals( 0f, am.T()[0], 1e-6f );
		
		trainStage1(0,0, 0);
		assertEquals( 2, am.C() );
		assertEquals( 1f, am.w()[0][0], 1e-6f );
		assertEquals( 1f, am.w()[0][1], 1e-6f );
		assertEquals( 0f, am.w()[1][0], 1e-6f );
		assertEquals( 0f, am.w()[1][1], 1e-6f );
		
		
	}
	
	/**
	 * Given two category boxes, does system choose closer box.
	 */
	public void testPicksCloserBox() {
		
		am = createInstance(1,2);
		am.setUp();
		
		final float u0 = 0.0f, v0 = 0.3f;
		final float u1 = 0.6f, v1 = 0.9f;
		
		trainStage1( u0, 0 ); //Create R0
		trainStage1( v0, 0 ); //Extend right
		trainStage1( u1, 1 ); //Create R1
		trainStage1( v1, 1 ); //Extend right
		
		final float[][] w = am.w();
		assertEquals( 2, w.length );
		assertEquals( u0, w[0][0], 1e-7f);
		assertEquals( 1f-v0, w[0][1], 1e-7f);
		assertEquals( u1, w[1][0], 1e-7f);
		assertEquals( 1f-v1,w[1][1], 1e-7f);

		assertEquals(1,test(0.5f)); //Since u1 is closer than v0
		
	}
	
	/**
	 * Tests that, given an input inside two category boxes, picks the smaller box,
	 * and given training to the contrary, match-tracking results in a new box.
	 */
	public void testPicksSmaller() {
		
		am = createInstance(1,2);
		am.setUp();
		
		final float u0 = 0.3f, v0 = 0.8f;
		final float u1 = 0.4f, v1 = 0.7f;
		
		trainStage1( u0, 0 ); //Create R0
		trainStage1( v0, 0 ); //Extend right
		trainStage1( u1, 1 ); //Create R1
		trainStage1( v1, 1 ); //Extend right
		
		final float a = 0.5f;
		final int K = 0;
		
		assertEquals( 1, test(a) ); //Picks smaller box (R1)
		trainStage1( a, K ); //Training should force creating of point box R2
		
		final float[][] w = am.w();
		assertEquals( 3, w.length );
		
		//Initial two boxes
		assertEquals( u0, w[0][0], 1e-7f);
		assertEquals(  1f-v0,   w[0][1], 1e-7f);
		assertEquals( u1, w[1][0], 1e-7f);
		assertEquals(  1f-v1,   w[1][1], 1e-7f);
		
		//New point box
		assertEquals( a,  w[2][0], 1e-7f); 
		assertEquals(  1f-a,    w[2][1], 1e-7f);
	}
	
	/**
	 * Tests that F2 and Fab ties are broken correctly.
	 */
	public void testTieBreaker() {
		
		am = createInstance(1,2);
		am.setUp();
		
		final float u0 = 0.0f, v0 = 0.4f;
		final float u1 = 0.6f, v1 = 1.0f;
		
		final LabeledInputsImpl training = new LabeledInputsImpl(4,1,2);
		training.add( u0, 0 ); //Create R0
		training.add( v0, 0 ); //Extend right
		training.add( u1, 1 ); //Create R1
		training.add( v1, 1 ); //Extend right
		
		am.trainStage1(training);		
		//R0 and R1 are same size and equidistant from a=0.5
		
		//First make sure there's no tie break elswhere.
		assertEquals( 0, test(0.1f) );
		float[] T = am.T();
		assertTrue( T[0] > T[1]+am.getTieThreshold() );//No tie break here
					
		//Now test tie breaking
		
		final Random r = new Random(am.getTieBreakerSeed());
		
		am.setEpsilon(-am.getEpsilon()); //MT+, just for this test.
		final float a = 0.5f;
		final int K = 0;
		trainStage1( a, K );
		final float[][] w = am.w();
		if (r.nextInt(2)==0) {
			// R0 won, and so extended right.
			assertEquals( 1f-a, w[0][1], 1e-6f);
		}
		else {
			// R1 won, and so a new point box is created
			assertEquals( 3, w.length );
		}
		
		am.setTieBreakerSeed(am.getTieBreakerSeed()+1);
		r.setSeed(am.getTieBreakerSeed());
		am.resetLearning();
		am.trainStage1( training );
		
		assertEquals(r.nextInt(2),test(a));
		assertEquals(r.nextInt(2),test(a));
		
	}
	
	/**
	 * Tests that match tracking accepts only a smaller box on second try.
	 */
	public void testMatchTracking() {
		
		am = createInstance(1,2);
		am.setUp();
		
		final float u0 = 0.3f, v0 = 0.8f;
		final float u1 = 0.4f, v1 = 0.7f;
		
		trainStage1( u0, 0 ); //Create R0
		trainStage1( v0, 0 ); //Extend right
		trainStage1( u1, 1 ); //Create R1
		trainStage1( v1, 1 ); //Extend right
		
		// Test that smaller second-choice box is allowed to expand
		final float a1 = 0.25f;
		final int K1 = 1;
		
		assertEquals( 0, test(a1) ); //In larger box
		trainStage1(a1,K1); //Should allow smaller box to expand, since |R2+a|<|R1|
		
		float[][] w = am.w();
		assertEquals(2, w.length);
		assertEquals( u0, w[0][0], 1e-7f);
		assertEquals(  1f-v0,   w[0][1], 1e-7f);
		assertEquals( a1,  w[1][0], 1e-7f); //Expanded left
		assertEquals(  1f-v1,   w[1][1], 1e-7f);
	
		//Test that larger second-choice box won't pass MT vigilance
		final float a2 = 0.2f;
		final int K2 = 0;
		assertEquals( 1, test(a2) ); //In smaller box
		trainStage1(a2,K2); //Shouldn't allow larger box to expand.
		w = am.w();
		assertEquals(3, w.length); //New box created
		assertEquals( u0, w[0][0], 1e-7f);
		assertEquals(  1f-v0,   w[0][1], 1e-7f);
		assertEquals( a1,  w[1][0], 1e-7f);
		assertEquals(  1f-v1,   w[1][1], 1e-7f);
		//New point box
		assertEquals( a2,  w[2][0], 1e-7f);
		assertEquals(  1f-a2,   w[2][1], 1e-7f);
			
	}
	
	/**
	 * Tests that MT- allows coding of inconsistent cases.
	 * I.e., when you try to train on a pair (a,2) where there is already a
	 * point-box for class 0 at a, the system should allow a new point box for
	 * class 1 to be created on top of the original point box.
	 */
	public void testMTMinus() {
		
		am = createInstance(1,2);
		am.setUp();
		final float a = 0.5f;
		
		trainStage1( a, 0 ); //Creates point box R0 for class 0 at a
		trainStage1( a, 1 ); //Allows point box R1 for class 1 to be created on top of R0
		
		//assert that R1=R2=a
		final float[][] w = am.w();
		assertEquals( 2, w.length );
		assertEquals( a, w[0][0], 0f );
		assertEquals(  1f-a,   w[0][1], 0f );
		assertEquals( a, w[1][0], 0f );
		assertEquals(  1f-a,   w[1][1], 0f );
		final int[] kappa = am.kappa();
		assertEquals(0,kappa[0]);
		assertEquals(1,kappa[1]);
		
		//Test IG-CAM properly handles point-boxes
		
		final float[] pK = am.testDistribution(new LabeledInput.Impl(new float[]{a},-1)); 
		assertEquals(new float[]{0.5f,0.5f},pK,1e-6f);
		
	}
	
	/**
	 * Test picks RJ if a in RJ, regardless of neighboring boxes
	 */
	public void testIGCAM() {
		
		am = createInstance(1,2);
		am.setUp();
		
		trainStage1( 0.1f, 0); //Create R0
		trainStage1( 0.3f, 0); //Expand right
		trainStage1( 0.4f, 1); //Create R1
		trainStage1( 0.6f, 1); //Expand right
		trainStage1( 0.7f, 0); //Create R2
		trainStage1( 0.9f, 0); //Expand right

		final float[][] w = am.w();
		assertEquals( 3, w.length );
		assertEquals( new float[]{0.1f,0.7f}, w[0], 1e-6f );
		assertEquals( new float[]{0.4f,0.4f}, w[1], 1e-6f );
		assertEquals( new float[]{0.7f,0.1f}, w[2], 1e-6f );
		final int[] kappa = am.kappa();
		assertEquals(0,kappa[0]);
		assertEquals(1,kappa[1]);
		assertEquals(0,kappa[2]);
	
		final float[] pK = am.testDistribution(new LabeledInput.Impl(new float[]{0.5f},-1)); 
		assertTrue( pK[1]-pK[0] > 1e-6f ); //R1 should win and not just by tie breaker
	
	}
	
	/**
	 * All else equal, a node with existing feature should beat out a node missing that feature.
	 * (Third detailed illustration for publication)
	 */
	public void testExistingBeatsMissing() {
		
		am = createInstance(2,2);
		am.setUp();
		
		final float a = .5f;
		final int K = 0;
		for (final float margin : new float[]{ 0.0f, 0.1f, 0.4f }) {
			try {
				
				//Create R1, fully featured
				trainStage1( a-margin,a-margin, 0 );
				trainStage1( a+margin,a+margin, 0 );
				
				//Create R2, missing second feature
				trainStage1( a-margin, NaN, 1 );
				trainStage1( a+margin, NaN, 1 );
				
				//R1 should win, and not just by a tie breaker
				assertEquals( K, test(a,a) );
				final float[] T = am.T();
				assertTrue( T[0] > T[1]+am.getTieThreshold() );
			
			} catch (AssertionFailedError e) {
				e.initCause(new UnitTestFailureContext("margin=%f",margin));
				throw e;
			}
		}
		
	}
	
	/**
	 * Verify that IG-CAM rule still works (yJ==1 if a in RJ) even with undercomitted nodes
	 * (First detailed illustration for publication)
	 */
	public void testInWinsWithMissing1() {
		
		am = createInstance(2,2);
		am.setUp();
		
		//Create R1 in the middle of dim 1, missing dim 2.
		final float u1 = 0.4f, v1 = 0.6f;
		trainStage1( u1,NaN, 0 );
		trainStage1( v1,NaN, 0 );
		
		//Create R2 and R3 at extremes of dim1, also missing dim 2.
		final float u2 = 0.1f, v2 = 0.3f; 
		trainStage1( v2,NaN, 1 );
		trainStage1( u2,NaN, 1 );
		final float u3 = 0.7f, v3 = 0.9f; 
		trainStage1( u3,NaN, 1 );
		trainStage1( v3,NaN, 1 );
		
		//Make sure training was correct.
		final float[][] w = am.w();
		assertEquals(new float[]{ u1, 1f, 1f-v1, 1f },w[0],1e-6f);
		assertEquals(new float[]{ u2, 1f, 1f-v2, 1f },w[1],1e-6f);
		assertEquals(new float[]{ u3, 1f, 1f-v3, 1f },w[2],1e-6f);
		assertEquals(new int[] { 0, 1, 1 },am.kappa());
		
		//Verify that IG-CAM results in R1 winning => K'=0
		assertEquals( 0, test(0.5f,0.5f) );		
		
	}
	
	/**
	 * If input is "in" a node missing a feature and not in a fully committed node,
	 * The first node should win. (Second detailed illustration for publication)
	 */
	public void testInWinsWithMissing2() {
		
		am = createInstance(2,2);
		am.setUp();
		
		//Create R1 to the left of dim1, fully committed.
		final float[] u1 = {0.1f,0.3f}, v1 = {0.3f,0.6f};
		trainStage1( u1[0],u1[1], 1 );
		trainStage1( v1[0],v1[1], 1 );
		
		//Create R2 to the right of dim1, uncommitted in dim2.
		final float u2 = 0.6f, v2 = 0.9f; 
		trainStage1( u2,NaN, 0 );
		trainStage1( v2,NaN, 0 );
		
		//Make sure training was correct.
		final float[][] w = am.w();
		assertEquals(new float[]{ u1[0],u1[1], 1f-v1[0],1f-v1[1] }, w[0], 1e-6f);
		assertEquals(new float[]{ u2,   1f,      1f-v2,  1f  }, w[1], 1e-6f);
		assertEquals(new int[] { 1, 0 },am.kappa());
		
		//Verify that IG-CAM results in R2 winning => K'=0
		assertEquals( 0, test(0.75f,0.5f) );		
	}
	
	/**
	 * d(Rj,a) should be scaled by the number of existing features,
	 * such that a R2 must be twice as close as R1 if it's half as committed.
	 * (Third detailed illustration for publication)
	 */
	public void testDistanceScaledByExisting() {
		
		am = createInstance(2,2);
		am.setUp();
		
		//Create R1 to the left of dim1, fully committed and perfect match in dim2.
		final float[] u1 = {0.1f,0.5f}, v1 = {0.3f,0.5f};
		final int K1 = 1;
		trainStage1( u1[0],u1[1], K1 );
		trainStage1( v1[0],v1[1], K1 );
		
		//Create R2 to the right of dim1, uncommitted in dim2.
		final float u2 = 0.7f, v2 = 0.9f; 
		final int K2 = 0;
		trainStage1( u2,NaN, K2 );
		trainStage1( v2,NaN, K2 );
		
		//Make sure training was correct.
		final float[][] w = am.w();
		assertEquals(new float[]{ u1[0],u1[1], 1f-v1[0],1f-v1[1] }, w[0], 1e-6f);
		assertEquals(new float[]{ u2,   1f,     1f-v2,  1f  }, w[1], 1e-6f);
		assertEquals(new int[] { K1, K2 },am.kappa());
		
		//Equal distance in x, R1 should win
		assertEquals( K1, test( 0.5f, 0.5f ) );
		
		//>Twice as close to R2 in x, R2 should win
		assertEquals( K2, test( 0.65f, 0.5f ) );
		
		//==Twice as close to R2 in x, R2 should win, 
		assertEquals( K2, test( 0.6f, 0.5f ) );
		final float[] T = am.T();
		assertTrue( T[1] > T[0]+am.getTieThreshold() ); //and not just by tie breaker
			
	}
	
	/**
	 * Make sure learning extends RJ only part way using slow learning.
	 */
	public void testSlowSupervisedLearning() {
		am = createInstance(1,1);
		am.setBeta1(0.25f);
		am.setUp();
		
		trainStage1( 0.3f, 0 ); //Fast-commit
		assertEquals( new float[]{ 0.3f, 0.7f }, am.w()[0], 1e-6f );
		
		trainStage1( 0.7f, 0 ); //Slow recode
		assertEquals( new float[]{ 0.3f, 0.6f }, am.w()[0], 1e-6f );
	}
	
	/**
	 * When SSL input is in only one box RJ, RJ should learn a fair amount.
	 */
	public void testConfidentSelfSupLearning() {
		
		am = createInstance(2,2);
		am.setBeta2(1f);
		am.setUp();
		
		trainStage1( 0.2,NaN, 0);
		trainStage1( 0.6,NaN, 0);
		trainStage1( 0.4,NaN, 1);
		trainStage1( 0.8,NaN, 1);
		
		assertEquals( 2, am.C() );
		assertEquals( new int[]{ 0, 1 }, am.kappa() );
		assertEquals( new float[]{ 0.2f, 1f, 0.4f, 1f }, am.w()[0], 1e-6f );
		assertEquals( new float[]{ 0.4f, 1f, 0.2f, 1f }, am.w()[1], 1e-6f );
		
		trainStage2( 0.3f, 0.25f );
		//Point was in R0 only, so should get most of the learning.
		
		float T0 = 1f/0.505f + 0.01f*(1f-0.4f);
		assertEquals( T0, am.T()[0], 1e-6f );
		float T1 = (1f-.1f)/0.505f + 0.01f*(1f-0.4f);
		assertEquals( T1, am.T()[1], 1e-6f );
		float y0 = (float) Math.pow(2f-T0,-2);
		float y1 = (float) Math.pow(2f-T1,-2);
		y0 /= (y0+y1);
		assertEquals( y0, am.y()[0], 0.01f );
		
		float w01 = 0.25f + (1f-y0);
		float w01c = 0.75f + (1f-y0);
		float[] w = am.w()[0];
		//Nothing changed in dim1
		assertEquals( 0.2f, w[0], 1e-6f );
		assertEquals( 0.4f, w[2], 1e-6f );
		//Assert learning within 1% of expected
		assertEquals( w01, w[1], 0.01f*w01 );
		assertEquals( w01c, w[3], 0.01f*w01c );
		
		//No change in R1
		assertEquals( new float[]{ 0.4f, 1f, 0.2f, 1f }, am.w()[1], 1e-6f );
	}
	
	/** Make sure cloning works */
	public void testClone() {
		am = createInstance(2,2);
		am.setUp();
		
		final float[] a = { 0.1f, 0.5f };
		final int K = 1;
		trainStage1( a[0],a[1], K );
		assertEquals( K, test(a) );
		
		am = (SelfSupervisedARTMAP) am.clone();
		am.setUp();
		assertEquals( K, test(a) );
		
	}
	
	/**
	 * Make sure SS-AM doesn't allow bad inputs.
	 */
	public void testBadInputs() {
		
		am = createInstance(2,2);
		am.setUp();
		
		try {
			trainStage1( -1, 0, 1); //Input out of bounds
			fail();
		}
		catch(InputFeatureOutOfARTBoundsException e) { /*good*/	}
		
		try {
			trainStage1( 0.2, 0.3, 5); //Class out of bounds
			fail();
		}
		catch(ClassLabelOutOfBoundsException e) { /*good*/	}

		try {
			test( 0.3f, 1.1f ); //Input out of bounds
			fail();
		}
		catch(InputFeatureOutOfARTBoundsException e) { /*good*/	}
		
		try {
			test( 0.3f ); //Wrong number of features
			fail();
		}
		catch(InputFeaturesLengthMismatchException e) { /*good*/	}
		
		try {
			trainStage2( 0.3f, -1.1f ); //Input out of bounds
			fail();
		}
		catch(InputFeatureOutOfARTBoundsException e) { /*good*/	}
		
		//Allow NaNs
		trainStage1( NaN, 1f, 1 );
		trainStage2( NaN, 0.4f );
		test( NaN, 0f );		
		
	}
	
	protected SelfSupervisedARTMAP createInstance(final int M, final int L) {
		final SelfSupervisedARTMAP am = new SelfSupervisedARTMAP();
		am.setP(2);
		am.setEpsilon(-0.01f);
		am.setM(M);
		am.setL(L);
		return am;
	}
	
	/**
	 * Convenience method for training on input aThenK(1:end-1) and label K
	 * @param aThenK
	 */
	protected void trainStage1(final Number... aThenK) {
		final int M = am.getM();
		assert aThenK.length==M+1;
		final float[] a = new float[M];
		for (int i=0; i<M; i++) a[i] = aThenK[i].floatValue();
		final int K = aThenK[M].intValue();
		am.trainStage1( new LabeledInput.Impl( a, K ) );		
	}
	
	/**
	 * Convenience method for testing on input a
	 * @param aThenK
	 */
	protected int test(final float... a) {
		return am.test( new LabeledInput.Impl(a,-1) );
	}
	
	/**
	 * Convenience method for selfSup on input a
	 * @param aThenK
	 */
	protected void trainStage2(final float... a) {
		am.trainStage2( new LabeledInput.Impl(a,-1) );
	}
	
	
	
	/**
	 * @deprecated
	 * @param e
	 * @param a
	 */
	public static void assertEquals(Object e, Object a) {
		throw new UnsupportedOperationException();
	}
	
	public static void assertEquals( int[] expected, int[] actual  )
	{
		
		if ( expected==null )
			return;
		assertNotNull(actual);
		
		assertEquals( "Different dimension. Expected " 
				+ expected.length + " got " + actual.length + ".",
				expected.length, actual.length );
		
		final int N = expected.length;
		for (int i = 0; i<N; i++) {
			assertEquals( "Different values at ["+i+"].",
					expected[i], actual[i] );
		}
		
	}
	
	public static void assertEquals(
			float[] expected, float[] actual, float delta  )
	{
		
		assertEquals( "Different dimensions: ",
				expected.length, actual.length );
		
		final int N = expected.length;
		for (int i = 0; i<N; i++) {
			assertEquals( "Different values at ["+i+"]. ("+(actual[i]-expected[i])+")",
					expected[i], actual[i], delta );
		}
		
	}
}
