/**
 *	InputsImpl.java
 *
 *  Greg Amis
 *  Technology Lab
 *  Dept. of Cognitive & Neural Systems
 *  Boston University
 *  Oct 15, 2008
 *
 *  Licensing Policy
 *  This implementation is made available to the general public under the terms 
 *  of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
 *  (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
 *  As such, it is available free of charge, and may be copied, redistributed, or modified, 
 *  as long as any resulting work is distributed under the same terms of Copyleft.
 */
package edu.bu.cns.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

/**
 *	Inputs implementation backed by a List
 */
public class InputsImpl implements Inputs<Input> {

	
	private final int numFeatures;
	private final List<Input> inputs;
	
	
	/**
	 * Creates an empty ArrayList-backed instance.
	 * @param initCapacity
	 * @param numFeatures
	 */
	public InputsImpl(final int initCapacity, final int numFeatures)
	{
		this.numFeatures = numFeatures;
		this.inputs = createList(initCapacity);		
	}
	
	/**
	 * Creates an instance based on the passed inputs.
	 * Each input must have the same number of features.
	 * 
	 * @param inputs 
	 */
	public InputsImpl(final float[][] inputs) {
		this.numFeatures = (inputs.length>0) ? inputs[0].length : 0;
		this.inputs = createList(inputs.length);
		
		for (int n=0; n<inputs.length; n++) {
			this.inputs.add( new Input.Impl( inputs[n] ) );
		}
	}

	/**
	 * 	Creates an instance backed by the passed list.
	 * @param list
	 */
	public InputsImpl(final List<Input> list) {
		this.numFeatures = (list.isEmpty()) ? 0 : list.get(0).features().length;
		this.inputs = list;
	}
	
	public float[][] features() {
		final float[][] features = new float[size()][numFeatures];
		int i = 0;
		for (final Input input : this)
			System.arraycopy(input.features(), 0, features[i++], 0, numFeatures);
		
		return features;
	}

	public Input get(final int i) { return inputs.get(i); }
	
	/**
	 * Convenience method for quickly adding an input
	 * @param features
	 */
	public void add(final Number... features) {
		InputFeaturesLengthMismatchException.assertMatchingNumberOfInputs(
				numFeatures, features.length );
		final float[] floatFeatures = new float[numFeatures];
		for (int i=0; i<numFeatures; i++) 
			floatFeatures[i] = features[i].floatValue();
		inputs.add( new Input.Impl(floatFeatures) );
	}
	
	public int getNumberOfFeatures() { return numFeatures;	}
	
	public Inputs<Input> createShuffledCopy(final Random r) {
		final List<Input> dest = createList(inputs);
		Collections.shuffle(dest,r);
		return new InputsImpl(dest);
	}

	public Inputs<Input> subsequence(int start, int end) {
		return new InputsImpl(inputs.subList(start, end));
	}
	
	/**
	 * Creates a new ArrayList
	 * @param initCapacity
	 * @return
	 */
	protected List<Input> createList(final int initCapacity)
	{
		return new ArrayList<Input>(initCapacity);
	}
	
	/**
	 * Creates a new ArrayList of size src.size(), adding all the elements of src.
	 * @param src
	 * @return
	 */
	protected final List<Input> createList( final Collection<Input> src )
	{
		final List<Input> list = createList(src.size());
		list.addAll(src);
		return list;
	}

	public int size() { return inputs.size();	}

	public Iterator<Input> iterator() { return inputs.iterator(); }


}
