/**
 *	LabeledInputsImpl.java
 *
 *  Greg Amis
 *  Technology Lab
 *  Dept. of Cognitive & Neural Systems
 *  Boston University
 *  Oct 15, 2008
 *
 *  Licensing Policy
 *  This implementation is made available to the general public under the terms 
 *  of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
 *  (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
 *  As such, it is available free of charge, and may be copied, redistributed, or modified, 
 *  as long as any resulting work is distributed under the same terms of Copyleft.
 */
package edu.bu.cns.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

/**
 *	LabeledInputs implementation backed by a List
 */
public class LabeledInputsImpl implements LabeledInputs {

	private static final long serialVersionUID = 7373019254298664960L;
	
	private final int numFeatures, numClasses;
	private final List<LabeledInput> inputs;
	
	
	/**
	 * Creates an empty ArrayList-backed instance.
	 * @param initCapacity
	 * @param numFeatures
	 * @param numClasses
	 */
	public LabeledInputsImpl(final int initCapacity, final int numFeatures, final int numClasses)
	{
		this.numFeatures = numFeatures;
		this.numClasses = numClasses;
		this.inputs = createList(initCapacity);		
	}
	
	/**
	 * Creates an instance based on the passed inputs and corresponding labels.
	 * Thus, the number of inputs must equal the number of labels,
	 * and each input must have the same number of features.
	 * 
	 * @param inputs 
	 * @param labels
	 * @param numClasses
	 */
	public LabeledInputsImpl(final float[][] inputs, final int[] labels, final int numClasses) {
		
		if (inputs.length!=labels.length)
			throw new InputsLabelsLengthMismatchException(inputs.length,labels.length);
		
		this.numFeatures = (inputs.length>0) ? inputs[0].length : 0;
		this.numClasses = numClasses;
		this.inputs = createList(inputs.length);
		
		for (int n=0; n<inputs.length; n++) {
			this.inputs.add( new LabeledInput.Impl( inputs[n],labels[n] ) );
		}
	}

	/**
	 * 	Creates an instance backed by the passed list.
	 * @param list
	 * @param numClasses
	 */
	public LabeledInputsImpl(final List<LabeledInput> list, final int numClasses) {
		this.numFeatures = (list.isEmpty()) ? 0 : list.get(0).features().length;
		this.numClasses = numClasses;
		this.inputs = list;
	}
	
	public float[][] features() {
		final float[][] features = new float[size()][numFeatures];
		int i = 0;
		for (final LabeledInput input : this)
			System.arraycopy(input.features(), 0, features[i++], 0, numFeatures);
		
		return features;
	}

	public int[] labels() {
		final int[] labels = new int[size()];
		int i = 0;
		for (final LabeledInput input : this)
			labels[i++] = input.label();
		return labels;
	}
	
	public LabeledInput get(final int i) { return inputs.get(i); }
	
	/**
	 * Convenience method for quickly adding an input
	 * @param featuresThenLabel First elements are the input, last element is label
	 */
	public void add(final Number... featuresThenLabel) {
		InputFeaturesLengthMismatchException.assertMatchingNumberOfInputs(
				numFeatures, featuresThenLabel.length-1 );
		final float[] features = new float[numFeatures];
		for (int i=0; i<numFeatures; i++) 
			features[i] = featuresThenLabel[i].floatValue();
		final int label = featuresThenLabel[numFeatures].intValue();
		inputs.add( new LabeledInput.Impl(features,label) );
	}
	

	/** 
	 *  Creates a new LabeledInputs, with each class downsampled 
	 *  such that no class has more than cap points 
	 */
	public LabeledInputs downsampleToLabelCap( final Random r, final int cap) {
		
		List<LabeledInput> dest;
		
		if (cap>=size()) {
			//No downsampling necessary, just keep all the inputs
			dest = createList(inputs);
		}
		else {
			//Determine the number of points we want to keep per class
			final int[] ptsPerClass = new int[getNumberOfClasses()];
			for (final LabeledInput li : this)
				ptsPerClass[li.label()]++;
		
			//And the total number of points.
			int len = 0;
			for (final int nk : ptsPerClass)
				len += (nk<cap) ? nk : cap;
		
			final List<LabeledInput> shuffleSrc = createList(inputs);
			Collections.shuffle(shuffleSrc,r);
			dest =  createList(len);
			
			//Reuse ptsPerClass as a place to keep a running count
			//of how many points have been added for each class
			Arrays.fill(ptsPerClass,0);
		
			for ( final LabeledInput li : shuffleSrc ) {
				if (ptsPerClass[li.label()]<cap) {
					//Cap not yet reached. Add the point.
					dest.add( li );
					ptsPerClass[li.label()]++;
				}
			}
		
			assert dest.size()==len;
		}
		
		//Must be reshuffled, otherwise, the end of the list 
		// has all the bigger classes.
		Collections.shuffle(dest, r);
		
		return new LabeledInputsImpl(dest,numClasses);
		
	}

	public int getNumberOfClasses() { return numClasses;	}

	public int getNumberOfFeatures() { return numFeatures;	}
	
	public LabeledInputs createShuffledCopy(final Random r) {
		final List<LabeledInput> dest = createList(inputs);
		Collections.shuffle(dest,r);
		return new LabeledInputsImpl(dest,numClasses);
	}

	public LabeledInputs subsequence(int start, int end) {
		return new LabeledInputsImpl(inputs.subList(start, end),numClasses);
	}
	
	public List<LabeledInput> list() { return inputs; }
	
	/**
	 * Creates a new ArrayList
	 * @param initCapacity
	 * @return
	 */
	protected List<LabeledInput> createList(final int initCapacity)
	{
		return new ArrayList<LabeledInput>(initCapacity);
	}
	
	/**
	 * Creates a new ArrayList of size src.size(), adding all the elements of src.
	 * @param src
	 * @return
	 */
	protected final List<LabeledInput> createList( final Collection<LabeledInput> src )
	{
		final List<LabeledInput> list = createList(src.size());
		list.addAll(src);
		return list;
	}

	public int size() { return inputs.size();	}

	public Iterator<LabeledInput> iterator() { return inputs.iterator(); }


}
