/**
 *	Pow.java
 *
 *  Greg Amis
 *  Technology Lab
 *  Dept. of Cognitive & Neural Systems
 *  Boston University
 *  Nov 17, 2008
 *
 *  Licensing Policy
 *  This implementation is made available to the general public under the terms 
 *  of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
 *  (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
 *  As such, it is available free of charge, and may be copied, redistributed, or modified, 
 *  as long as any resulting work is distributed under the same terms of Copyleft.
 */
package edu.bu.cns.util;

/**
 *	More efficient version of Math.pow. Best for situations 
 *	where y^p is calculated for p = 1 or p = a small integer.
 *	This should not be used for high precision or high-exponent situations.
 *	Output error (relative to Math.pow) is in general < expected*10^-6
 */
public abstract class Pow {
	
	

	/**
	 * Returns the best Pow strategy for the passed exponent p.
	 * @param p
	 * @return
	 */
	public static Pow createPow(final Number p) {
		
		final int p_int = p.intValue();
		if (p_int==p.floatValue()) { 
			if (p_int==0)
				return new ZeroPow();
			return p_int>0 ? new IntPow(p_int) : new NegIntPow(p_int);
		}
		
		return new MathPow(p.doubleValue());
	}
	
	/**
	 * Returns base^p
	 * @param base
	 * @return
	 */
	public abstract float calc(final float base);
	
	/**
	 * 
	 * @return exponent p
	 */
	public abstract Number p();
	
	/**
	 * Pass through implementation for when exponent is 0. base^0=base
	 */
	private static class ZeroPow extends Pow {
		@Override
		public final float calc(final float base) { return 1f; }
		@Override
		public final Number p() { return 0; }
	}
	
	/**
	 * For small-integer exponents, calcs y=x^p by looping through y*=x p times.
	 */
	private static class IntPow extends Pow {
		private final int p;
		private IntPow(final int p) { this.p = p; }
		
		@Override
		public float calc(final float base) {
			float ret = base;
			for (int pp=1; pp<p; pp++) ret *= base;
			return ret;
		}
		@Override
		public Number p() { return p; }
	}
	
	/**
	 * Handles small negative integer exponents,
	 * calcs y=x^p by setting b=1/x then looping through y*=b -p times.
	 */
	private static class NegIntPow extends IntPow {
		public NegIntPow(final int p) { super(-p); }
		@Override
		public float calc(final float base) { return super.calc(1f/base); }
		@Override
		public Number p() { return -super.p().intValue();	}
	}
	
	/**
	 * Calls Math.pow(base,p) for cases where p is not an integer.
	 */
	private static class MathPow extends Pow {
		private final double p;
		private MathPow(final double p) { this.p = p; }
		@Override
		public float calc(final float base) { return (float) Math.pow(base,p); }
		@Override
		public Number p() { return p; }
	}
	
	@Override
	public String toString() { 
		return String.format("(.)^%d",p());
	}
}
